package com.forumtopics.bob;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import com.businessobjects.sdk.plugin.desktop.customrole.ICustomRole;
import com.businessobjects.sdk.plugin.desktop.customrole.IRoleRight;
import com.crystaldecisions.sdk.exception.SDKException;
import com.crystaldecisions.sdk.exception.SDKException.PluginNotFound;
import com.crystaldecisions.sdk.exception.SDKException.PluginNotFoundAtCMS;
import com.crystaldecisions.sdk.framework.CrystalEnterprise;
import com.crystaldecisions.sdk.framework.IEnterpriseSession;
import com.crystaldecisions.sdk.framework.ISessionMgr;
import com.crystaldecisions.sdk.occa.infostore.CeSecurityCUID;
import com.crystaldecisions.sdk.occa.infostore.IInfoObject;
import com.crystaldecisions.sdk.occa.infostore.IInfoObjects;
import com.crystaldecisions.sdk.occa.infostore.IInfoStore;
import com.crystaldecisions.sdk.occa.pluginmgr.IPluginInfo;
import com.crystaldecisions.sdk.occa.pluginmgr.IPluginMgr;
import com.crystaldecisions.sdk.plugin.desktop.program.IProgramBase;
import com.eaio.util.Resource;

/**
 * Lists granted and denied rights in all custom access levels in a BOE environment.  1 argument, output file name, is required to run this class as
 * a program object
 * @author Ryan Gordon
 *
 */
public class RightsLister implements IProgramBase{

	// Plugins that are not displayed in the CMC
	final static String HIDDEN_OBJS = "secEnterprise,Service,Install,ServiceContainer,EnterpriseNode,ServiceCategory,secWindowsNT,secWinAD,secLDAP," +
			"WebiServerAdmin,Testapp,SystemObjectsFolder,Smtp,SSOAdmin,SLMetaDataInfoObject,ScopeBatch,RepositoryPromptGroup,ReportAppServerAdmin," +
			"PluginAction,PageServerAdmin,Overload,MetaData.MetaDataRepositoryInfo,MetaData.ReportCommand,MetaData.MetaDataTextObject," +
			"MetaData.MetaDataPictureObject,MetricDescriptions,MetaData.DynamicDataConnection,MetaData.DataTable,MetaData.MetaDataCustomFunction," +
			"MetaData.DataFoundation,MetaData.DataDBField,MetaData.DataCommandTable,MetaData.BusinessView,MetaData.BusinessParameterField," +
			"MetaData.BusinessFormulaField,MetaData.BusinessFilter,MetaData.BusinessField,MetaData.DataProcedure,MetaData.BusinessElement," +
			"MetaData.BusinessCompositeFilter,JobServerAdmin,InfoObject,InfoObjectsFolder,FullClientProcAdmin,FullClientCacheAdmin,Ftp,FileServerAdmin," +
			"EventServerAdmin,DocProcessingProcAdmin,DiskUnmanaged,DocProcessingCacheAdmin,Destination,DeltaStore,CMSAdmin,CsContainerAdmin," +
			"ClientActionUsage,ClientActionSet,ClientAction,Wizard,AppObjectsFolder,CacheServerAdmin,AuditEventInfo,AuditAdmin,M13NTarget," +
			"M13NTargetGroup,Materialization,PMModelM13N,Managed,Connection,JavaScheduling";
	
	// Built-in access levels
	private final static Set<String> SYSTEM_ROLE_CUIDS = new HashSet<String>();
	static{
		SYSTEM_ROLE_CUIDS.add(CeSecurityCUID.Role.FULL_CONTROL);
		SYSTEM_ROLE_CUIDS.add(CeSecurityCUID.Role.SCHEDULE);
		SYSTEM_ROLE_CUIDS.add(CeSecurityCUID.Role.VIEW);
		SYSTEM_ROLE_CUIDS.add(CeSecurityCUID.Role.VIEW_ON_DEMAND);
	}

	// Default category and rights collection name for rights that do not come from a plugin
	final static String DEFAULT = "General";
	
	// Default locale for rights description
	final static Locale LOC = Locale.getDefault();
	
	// Need default constructor explicitly defined when running as program object
	public RightsLister(){}

	@Override
	public void run(IEnterpriseSession enterpriseSession, IInfoStore infoStore, String[] args) throws SDKException {
		checkParams(args);
		IInfoObjects roles = getAccessLevels(infoStore);
		printAccessLevelContents(infoStore, roles, new File(args[0]));		
	}

	/**
	 * Checks the command line arguments for required args.  {@code FileName} is the only required argument
	 * @param args
	 */
	private void checkParams(String[] args){
		if (args.length != 1)
			throw new IllegalArgumentException("An output file name was not specified.  Exiting...");
	}

	public static void main(String[] args) throws Throwable{
		// When running from the command line, 5 args are needed:  connection info and output file path.  Error out otherwise
		if (args.length != 5){
			System.err.println("Incorrect number of arguments supplied.  Usage: <user> <pass> <CMS> <authentication> <output_file>");
			throw new IllegalArgumentException("Incorrect command line args");
		}

		ISessionMgr sm = CrystalEnterprise.getSessionMgr();
		IEnterpriseSession es = sm.logon(args[0], args[1], args[2], args[3]);
		IInfoStore infoStore =(IInfoStore)es.getService("InfoStore");
		
		try{
			new RightsLister().run(es, infoStore, new String[]{args[4]});
			System.out.println("Complete");
		}
		finally{
			Resource.insteadOfClose(es.getClass(), "logoff");
			Resource.close(es);
		}

	}

	/**
	 * Loops through all access levels and outputs explicitly granted and denied rights to System.out and to the file represented by {@code out}
	 * @param infoStore
	 * @param roles
	 * @param out
	 * @throws SDKException
	 */
	private void printAccessLevelContents(IInfoStore infoStore, IInfoObjects roles, File out) throws SDKException{
		// Used to write to the output file
		PrintWriter w = null;
		try{
			w = new PrintWriter(new FileWriter(out));
			
			String header = "Access Level\tRights Collection\tRight\tCategory\tIsGranted\tScope";
			System.out.println(header);
			w.println(header);

			// Get all applicable plugins for category and collection attributes
			Map<String, IPluginInfo> pluginsByKind = getAllPluginsByKind(infoStore);

			for (Object o : roles) {
				ICustomRole role = (ICustomRole)o;
				
				// Don't process built-in access levels
				if (SYSTEM_ROLE_CUIDS.contains(role.getCUID()))
					continue;
				
				for (Object objRR : role.getRoleRights()) {
					IRoleRight rr = (IRoleRight)objRR;
					String kind = rr.getKind();
					String aKind = rr.getApplicableKind();
					String coll = DEFAULT;
					String cat = DEFAULT;

					// Determine how to lookup the plugin source for this right.  This can be the plugin kind or the applicable type if the right is
					// a general system right such as "add objects to the folder"
					String key = (kind.length() > 0) ? kind : aKind;

					if (key.length() > 0){
						IPluginInfo pi = pluginsByKind.get(key);
						try{
							coll = pi.getTypeCollection(LOC);
							cat = pi.getName();
						}
						catch(NullPointerException e){
							// This right is either a general right (i.e. kind="" and aKind="Any") or the right belongs to a category that the CMC
							// does not expose.  If the latter, go on to the next right.  
							if (!aKind.equalsIgnoreCase("Any"))
								continue;
						}					
					}

					String output = String.format("%s\t%s\t%s\t%s\t%s\t%s", 
							role.getTitle(),
							coll,
							rr.getDescription(LOC),
							cat,
							rr.getGranted(),
							rr.getScope()
					);
					System.out.println(output);
					w.println(output);
				}
			}
		}
		catch(IOException e){
			throw new RuntimeException(e);
		}
		finally{
			Resource.beforeClose(w.getClass(), "flush");
			Resource.close(w);
		}
	}

	/**
	 * Queries the system for all custom access levels
	 * @param infoStore
	 * @return
	 * @throws SDKException
	 */
	private IInfoObjects getAccessLevels(IInfoStore infoStore) throws SDKException{

		// Probably should not use 'select *' here
		String query = "select si_role_rights, si_cuid from ci_systemobjects where si_kind='customrole'";
		return infoStore.query(query);
	}

	/**
	 * PluginInfo logic extracted from CmcAppActions web app and modified to work as needed 
	 * @param infoStore
	 * @return
	 * @throws SDKException
	 */
	private Map<String, IPluginInfo> getAllPluginsByKind(IInfoStore infoStore) throws SDKException{
		Set<String> exclPluginKinds = getExcludePluginTypes(infoStore);

		final IPluginMgr pMgr = infoStore.getPluginMgr();
		List<IPluginInfo> plugins = new ArrayList<IPluginInfo>(Arrays.asList(pMgr.getPlugins( IPluginMgr.Category.DESKTOP )));
		plugins.addAll(Arrays.asList(pMgr.getPlugins( "meta" )));

		Map<String, IPluginInfo> m = new HashMap<String, IPluginInfo>();
		for (IPluginInfo p : plugins) {
			String kind = p.getKind();

			// Don't add the plugin to the lookup list because it is not displayed in the CMC
			if (exclPluginKinds.contains(kind))
				continue;

			m.put(kind, p);
		}

		return m;
	}

	/**
	 * Generates a list of plugin types that are not displayed in the CMC.  This code was taken from the CmcAppActions web app and modified to function
	 * as needed
	 * @param infoStore
	 * @return
	 * @throws SDKException
	 */
	@SuppressWarnings("unchecked")
	private Set<String> getExcludePluginTypes(IInfoStore infoStore) throws SDKException {

		IPluginMgr pluginMgr = infoStore.getPluginMgr();
		Set<String> excludedPluginKinds = new HashSet<String>();

		// querying for all the applications under folder 99, which we DO want to display the rights for
		String query = "SELECT SI_KIND FROM CI_APPOBJECTS WHERE SI_PARENTID = 99"; 
		IInfoObjects objs = infoStore.query(query);

		String acceptableAppsStr = "";

		// creating part of the where clause for the next query. We want to filter out all the plugins
		// in the Application collection EXCEPT for the plugins under folder 99.
		for (Iterator<IInfoObject> it = objs.iterator(); it.hasNext();)
		{
			acceptableAppsStr += "AND SI_KIND NOT LIKE '";
			acceptableAppsStr += ( ( IInfoObject ) it.next() ).getKind();
			acceptableAppsStr += "' ";
		}

		// querying for all the application plugins that should be filtered out, ie, all the application
		// plugins we DON'T want to display the rights for.
		query = "SELECT SI_KIND FROM CI_SYSTEMOBJECTS WHERE SI_PARENTID=27 AND SI_RIGHTS_COLLECTION=2 " + acceptableAppsStr;
		objs = infoStore.query(query);

		// adding the hidden application plugins to the excluded plugins set.
		for (Iterator<IInfoObject> it = objs.iterator(); it.hasNext();)	{
			String pluginKind = null;
			String kind = it.next().getKind();
			pluginKind = getPluginKind( kind, pluginMgr );

			excludedPluginKinds.add( pluginKind );
		}

		// Get the list of hidden plugins.  Taken from cmc.kinds.hidden2 parameter in CmcAppActions web.xml file
		String pluginsStr = HIDDEN_OBJS;

		// adding the plugins from the web.xml list to the excluded plugins set.
		String[] stringArray = pluginsStr.split( "," );
		for (int i = 0; i < stringArray.length; i++){
			String kind = getPluginKind( stringArray[i], pluginMgr );
			excludedPluginKinds.add( kind );
		}

		return excludedPluginKinds; 
	}

	/**
	 * Gets the CeKind a plugin affects.  Code taken from CmcAppActions web app and modified to fit the needs of this class
	 * @param kind
	 * @param pluginMgr
	 * @return
	 * @throws SDKException
	 */
	private String getPluginKind( String kind, IPluginMgr pluginMgr ) throws SDKException {
		IPluginInfo info;
		try
		{
			info = pluginMgr.getPluginInfo( kind );
			if ( info != null )
			{
				return info.getKind();
			}
		} catch (PluginNotFound e)
		{
			// no-op
		} catch (PluginNotFoundAtCMS e)
		{
			// no-op

		} catch (SDKException e)
		{
			throw e;
		}
		// If we can't resolve the kind to a valid plugin type, just return the kind
		return kind;
	}
}
